package edu.berkeley.sbp.util;
import java.io.*;
import java.lang.reflect.*;

/** Random reflection-related utilities */
public final class Reflection {

    private static Object rebuild(Object o, Class c) {
        //System.out.println("rebuild " + o + " (a " + (o==null?null:o.getClass().getName()) + ") " + c);
        if (o==null || c.isAssignableFrom(o.getClass())) return o;
        if ((c == Character.class || c == Character.TYPE) && o instanceof String && ((String)o).length()==1) return new Character(((String)o).charAt(0));
        if (o instanceof Character && c == String.class) return o+"";
        if (o instanceof Object[]) {
            Object[] a = (Object[])o;
            if (c.isArray()) {
                Object[] ret = (Object[])Array.newInstance(c.getComponentType(), a.length);
                for(int i=0; i<ret.length; i++) {
                    Object o2 = rebuild(a[i], c.getComponentType());
                    //if (o2 != null) System.out.println("storing " + o2.getClass().getName() + " to " + c.getComponentType());
                    ret[i] = o2;
                }
                return ret;
            }
            if (c == String.class) {
                boolean ok = true;
                for(int i=0; i<a.length; i++)
                    if (a[i]==null || (!(a[i] instanceof Character) && !(a[i] instanceof String)))
                        ok = false;
                if (ok) {
                    StringBuffer s = new StringBuffer();
                    for(int i=0; i<a.length; i++)
                        s.append(a[i] instanceof Character
                                 ? (((Character)a[i]).charValue())+""
                                 : (String)a[i]
                                 );
                    return s.toString();
                }
            }
        } else if (c.isArray()) {
            Object[] ret = (Object[])Array.newInstance(c.getComponentType(), 1);
            ret[0] = o;
            return ret;
        } else {
            throw new Error("unable to cast " + o + " from " + o.getClass().getName() + " to " + c.getName());
        }
        return o;
    }

    public static Object[] newArray(Class c, int i) {
        return (Object[])Array.newInstance(c, i);
    }

    public static Object[] lub(Object[] argo) {
        if (argo==null) return null;
        Class c = null;
        for(Object o : argo) if (o != null) c = Reflection.lub(c, o.getClass());
        if (c==null) c = Object.class;
        Object[] ret = Reflection.newArray(c, argo.length);
        System.arraycopy(argo, 0, ret, 0, argo.length);
        return ret;
    }

    public static Class lub(Class a, Class b) {
        if (a==null) return b;
        if (b==null) return a;
        if (a==b) return a;
        if (a.isAssignableFrom(b)) return a;
        if (b.isAssignableFrom(a)) return b;
        return lub(b, a.getSuperclass());
    }

    /** a version of <tt>Class.forName</tt> that returns <tt>null</tt> instead of throwing an exception */
    public static Class forNameOrNull(String s) {
        try {
            return Class.forName(s); 
        } catch (ClassNotFoundException _) {
            return null;
        }
    }

    /** invoke method/constructor m on object o with arguments args, and perform reasonable coercions as needed */
    public static Object fuzzyInvoke(Object o, Member m, Object[] args) {
        try {
            Class[] argTypes = m instanceof Method ? ((Method)m).getParameterTypes() : ((Constructor)m).getParameterTypes();
            boolean ok = true;
            Object[] args2 = new Object[args.length];
            System.arraycopy(args, 0, args2, 0, args.length);
            args = args2;
            for(int i=0; i<args.length; i++) {
                Class goal = argTypes[i];
                Class actual = args[i] == null ? null : args[i].getClass();
                if (args[i] == null || goal.isAssignableFrom(actual)) continue;
                try {
                    args[i] = rebuild(args[i], goal);
                } catch (Error e) {
                    throw new Error(e.getMessage() + "\n   while trying to fuzzyInvoke("+m.getName()+")");
                }
            }
            // for debugging
            /*
            System.out.println("\ninvoking " + o + "." + m);
            for(int i=0; i<args.length; i++) {
                if (args[i] instanceof Object[]) {
                    System.out.print("  arg => " + zoo(args[i]));
                } else {
                    System.out.println("  arg => " + args[i] + (args[i]==null?"":(" (a " + args[i].getClass().getName() + ")")));
                }
            }
            */
            // FIXME: deal with the case where it is an inner class ctor
            return m instanceof Method ? ((Method)m).invoke(o, args) : ((Constructor)m).newInstance(args);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private static String zoo(Object o) {
        if (o==null) return "null";
        if (o instanceof Object[]) {
            String ret = "[ ";
            for(int j=0; j<((Object[])o).length; j++)
                ret += zoo(((Object[])o)[j]) + " ";
            return ret+"]";
        }
        return o+"";
    }

}
